"use strict";

var _toConsumableArray = require("babel-runtime/helpers/to-consumable-array")["default"];

var _slicedToArray = require("babel-runtime/helpers/sliced-to-array")["default"];

var _core = require("babel-runtime/core-js")["default"];

var _ = require("highland");

var _require = require("./util");

var denodeify = _require.denodeify;

var chalk = require("chalk");

var _require2 = require("docopt");

var docopt = _require2.docopt;

var read = denodeify(require("read"));
var readFile = denodeify(require("fs").readFile);
var writeFile = denodeify(require("fs").writeFile);
var yaml = require("js-yaml");
var yamlUpdate = require("yaml-update").edit;
var zipObject = require("lodash.zipobject");

var migrate = require("./");
var github = require("./github");

var doc = "\nUsage:\n  gogs-migrate [options]\n\nOptions:\n  -h, --help                Show help.\n  -V, --version             Show version.\n  -c, --config=<path>       Read options from given YAML configuration file.\n  --gogs-prefix=<prefix>    Gogs URL prefix (like `https://git.example.com`).\n  --gogs-token=<token>      Gogs access token (required, prompt otherwise).\n  --gogs-uid=<uid>          Gogs UID (required, prompt otherwise).\n  --github-prefix=<prefix>  GitHub API prefix (defaults to `https://api.github.com`).\n  --github-user=<user>      GitHub username to stream repositories from.\n  --github-token=<token>    GitHub user token to access private repositories.\n  --mirror                  Create mirror repositories.\n  --private                 Force private repositories (copy source visibility otherwise).\n  --with-forks              Include forked repositories.\n  --save                    Save migrated repositories to configuration file to avoid\n                            downloading them again the next time.\n".trim();

// Edit YAML in place (wrapper to avoid passing JSON).
var yamlEdit = function (source, data) {
  return yamlUpdate(source, JSON.stringify(data));
};

// Set properties on given object from a stream of name/value pairs.
var setProperties = function (object) {
  return _.pipeline(_.map(function (args) {
    return _.set.apply(_, _toConsumableArray(args));
  }), _.reduce(_core.Object.assign({}, object), function (opts, set) {
    return set(opts);
  }));
};

// Merge options and configuration file.
var config = function (opts) {
  return !opts["--config"] ? [opts] : readFile(opts["--config"], "utf8")
  // Keep original source and object for further usage.
  .map(function (source) {
    return [source, yaml.safeLoad(source)];
  }).map(function (_ref) {
    var _ref2 = _slicedToArray(_ref, 2);

    var source = _ref2[0];
    var data = _ref2[1];
    return _core.Object.assign({}, data, {
      "config-source": source,
      "config-object": data
    });
  })

  // Convert to options and extend `opts`.
  .flatMap(_.pairs).map(function (_ref) {
    var _ref2 = _slicedToArray(_ref, 2);

    var name = _ref2[0];
    var value = _ref2[1];
    return ["--" + name, value];
  }).pipe(setProperties(opts));
};

// Prompt for option if not defined.
var maybePrompt = function (opts) {
  return function (_ref) {
    var _ref2 = _slicedToArray(_ref, 2);

    var name = _ref2[0];
    var prompt = _ref2[1];
    return (opts[name] ? _([opts[name]]) : read(prompt)).map(function (value) {
      return [name, value];
    });
  };
};

var greyParens = function (text) {
  return text.replace("(", chalk.styles.grey.open + "(").replace(")", ")" + chalk.styles.grey.close);
};

var ask = function (question) {
  return "" + chalk.blue(">>") + " " + greyParens(question) + "" + chalk.styles.cyan.open;
};

// Prompt for required options if needed.
var prompt = function (opts) {
  return _([["--gogs-prefix", { prompt: ask("Gogs URL (example: `https://git.example.com`):") }], ["--gogs-token", { prompt: ask("Gogs token:") }], ["--gogs-uid", { prompt: ask("Gogs UID:") }]]).flatMap(maybePrompt(opts)).pipe(setProperties(opts));
};

// Parse options, apply configuration file and prompt missing values.
var opts = function (argv) {
  return _([docopt(doc, { argv: argv, version: require("./package").version })]).flatMap(config).flatMap(prompt)

  // Make sure `--ignore` is a hash.
  .map(function (opts) {
    return _core.Object.assign({}, opts, {
      "--ignore": zipObject(opts["--ignore"])
    });
  });
};

// Get repository stream from configured sources.
var repos = function (opts) {
  return _([

  // From GitHub.
  opts["--github-user"] && github.repos({
    user: opts["--github-user"],
    token: opts["--github-token"] || undefined,
    prefix: opts["--github-prefix"] || undefined
  })

  // Add sources here? Bitbucket, stdin, etc.

  ]).compact() // Remove falsy values.
  .sequence() // Merge inner streams.

  // Apply filters.
  .filter(opts["--with-forks"] ? function (repo) {
    return true;
  } : function (repo) {
    return !repo.fork;
  }).filter(function (repo) {
    return !(repo.name in opts["--ignore"]);
  });
};

// UI for regular migration stream.
var migrateStream = function (repoStream) {
  return repoStream.fork().map(function (_ref) {
    var repo = _ref.repo;
    return "" + chalk.green(">>") + " " + repo.name;
  }).errors(function (e, push) {
    if (e.repo) push(undefined, chalk.red(">> " + e.repo.name + " ") + e);else push(e);
  });
};

var saveConfig = function (opts, repos) {
  return [[opts["--config"], yamlEdit(opts["--config-source"], _core.Object.assign(opts["--config-object"], {
    ignore: _core.Object.keys(opts["--ignore"] || {}).concat(repos)
  }))]];
};

var saveNew = function (repos) {
  return _(read({
    prompt: ask("Where to save migrated repos (file will be overwritten)?")
  })).map(function (file) {
    return [file, yaml.safeDump({ ignore: repos })];
  });
};

// Save migrated repositories in configuration file at the end.
var saveStream = function (opts, repoStream) {
  return repoStream.fork().map(function (_ref) {
    var repo = _ref.repo;
    return repo.name;
  }).reduce([], function (memo, repo) {
    return memo.concat([repo]);
  })

  // Do nothing if there was no repositories.
  .filter(function (repos) {
    return repos.length;
  })

  // Append migrated repositories to ignore list.
  .flatMap(function (repos) {
    return opts["--config"] ? saveConfig(opts, repos) : saveNew(repos);
  }).flatMap(function (_ref) {
    var _ref2 = _slicedToArray(_ref, 2);

    var file = _ref2[0];
    var yaml = _ref2[1];
    return writeFile(file, yaml);
  }).map(function () {
    return "" + chalk.blue(">>") + " Configuration written!";
  })

  // Ignore repository errors (already handled).
  .errors(function (e, push) {
    if (!e.repo) push(e);
  });
};

// From fulfilled options, get repositories and migrate.
var main = function (opts) {
  var repoStream = repos(opts).flatMap(migrate({
    prefix: opts["--gogs-prefix"],
    token: opts["--gogs-token"],
    uid: opts["--gogs-uid"],
    mirror: opts["--mirror"],
    "private": opts["--private"]
  }));

  var stream = migrateStream(repoStream);

  return opts["--save"] ? _([stream, saveStream(opts, repoStream)]).merge() : stream;
};

module.exports = function (argv) {
  return opts(argv).flatMap(main).errors(function (e) {
    return console.error(e.stack || e);
  }).each(_.log);
};